"""Define hooks specifications and provide plugin manager"""

from __future__ import annotations

from typing import TYPE_CHECKING

from simplug import Simplug, SimplugResult
from xqute import JobStatus, Scheduler

from .defaults import ProcOutputType


if TYPE_CHECKING:  # pragma: no cover
    import signal
    from xqute import Xqute
    from .job import Job
    from .proc import Proc
    from .pipen import Pipen

plugin = Simplug("pipen")


@plugin.spec
def on_setup(pipen: Pipen) -> None:
    """Setup for plugins, primarily used for the plugins to
    setup some default configurations.

    This is only called once for all pipelines.

    Args:
        pipen: The Pipen object
    """


@plugin.spec
async def on_init(pipen: Pipen) -> None:
    """When the pipeline is initialized, and default configs are loaded

    Args:
        pipen: The Pipen object
    """


@plugin.spec
async def on_start(pipen: Pipen) -> None:
    """Right before the pipeline starts running.

    Process relationships are inferred.

    Args:
        pipen: The Pipen object
    """


@plugin.spec
async def on_complete(pipen: Pipen, succeeded: bool):
    """The the pipeline is completed.

    Args:
        pipen: The Pipen object
        succeeded: Whether the pipeline has successfully completed.
    """


@plugin.spec
def on_proc_create(proc: Proc):
    """Called Proc constructor when a process is created.

    Enables plugins to modify the default attributes of processes

    Args:
        proc: The Proc object
    """


@plugin.spec
async def on_proc_input_computed(proc: Proc):
    """Called after process input data is computed.

    Args:
        proc: The Proc object
    """


@plugin.spec
async def on_proc_script_computed(proc: Proc):
    """Called after process script is computed.

    The script is computed as a string that is about to compiled into a
    template.

    Args:
        proc: The Proc object
    """


@plugin.spec
async def on_proc_start(proc: Proc):
    """When a process is starting

    Args:
        proc: The process
    """


@plugin.spec(result=SimplugResult.TRY_ALL_FIRST_AVAIL)
def on_proc_shutdown(proc: Proc, sig: signal.Signals) -> None:
    """When pipeline is shutting down, by Ctrl-c for example.

    Return False to stop shutting down, but you have to shut it down
    by yourself, for example, `proc.xqute.task.cancel()`

    Only the first return value will be used.

    Args:
        pipen: The xqute object
        sig: The signal. `None` means a natural shutdown
    """


@plugin.spec
async def on_proc_done(proc: Proc, succeeded: bool | str) -> None:
    """When a process is done

    Args:
        proc: The process
        succeeded: Whether the process succeeded or not. 'cached' if all jobs
            are cached.
    """


@plugin.spec
async def on_job_init(job: Job):
    """When a job is initialized

    Args:
        job: The job
    """


@plugin.spec
async def on_job_queued(job: Job):
    """When a job is queued in xqute. Note it might not be queued yet in
    the scheduler system.

    Args:
        job: The job
    """


@plugin.spec(result=SimplugResult.TRY_ALL_FIRST_AVAIL)
async def on_job_submitting(job: Job) -> bool:
    """When a job is submitting.

    The first plugin (based on priority) have this hook return False will
    cancel the submission

    Args:
        job: The job

    Returns:
        False to cancel submission
    """


@plugin.spec
async def on_job_submitted(job: Job):
    """When a job is submitted in the scheduler system.

    Args:
        job: The job
    """


@plugin.spec
async def on_job_started(job: Job):
    """When a job starts to run in then scheduler system.

    Note that the job might not be running yet in the scheduler system.

    Args:
        job: The job
    """


@plugin.spec
async def on_job_polling(job: Job, counter: int):
    """When status of a job is being polled.

    Args:
        job: The job
    """


@plugin.spec(result=SimplugResult.TRY_ALL_FIRST_AVAIL)
async def on_job_killing(job: Job) -> bool:
    """When a job is being killed.

    The first plugin (based on priority) have this hook return False will
    cancel the killing

    Args:
        job: The job

    Returns:
        False to cancel killing
    """


@plugin.spec
async def on_job_killed(job: Job):
    """When a job is killed

    Args:
        job: The job
    """


@plugin.spec
async def on_job_succeeded(job: Job):
    """When a job completes successfully.

    Args:
        job: The job
    """


@plugin.spec
async def on_job_cached(job: Job):
    """When a job is cached.

    Args:
        job: The job
    """


@plugin.spec
async def on_job_failed(job: Job):
    """When a job is done but failed.

    Args:
        job: The job
    """


@plugin.spec(result=SimplugResult.ALL_AVAILS)
def on_jobcmd_init(job: Job) -> str:
    """When the job command wrapper script is initialized before the prescript is run

    This should return a piece of bash code to be inserted in the wrapped job
    script (template), which is a python template string, with the following
    variables available: `status` and `job`. `status` is the class `JobStatus` from
    `xqute.defaults.py` and `job` is the `Job` instance.

    For multiple plugins, the code will be inserted in the order of the plugin priority.

    The code will replace the `#![jobcmd_init]` placeholder in the wrapped job script.
    See also <https://github.com/pwwang/xqute/blob/master/xqute/defaults.py#L95>

    Args:
        job: The job object

    Returns:
        The bash code to be inserted
    """


@plugin.spec(result=SimplugResult.ALL_AVAILS)
def on_jobcmd_prep(job: Job) -> str:
    """When the job command right about to be run

    This should return a piece of bash code to be inserted in the wrapped job
    script (template), which is a python template string, with the following
    variables available: `status` and `job`. `status` is the class `JobStatus` from
    `xqute.defaults.py` and `job` is the `Job` instance.

    The bash variable `$cmd` is accessible in the context. It is also possible to
    modify the `cmd` variable. Just remember to assign the modified value to `cmd`.

    For multiple plugins, the code will be inserted in the order of the plugin priority.
    Keep in mind that the `$cmd` may be modified by other plugins.

    The code will replace the `#![jobcmd_prep]` placeholder in the wrapped job script.
    See also <https://github.com/pwwang/xqute/blob/master/xqute/defaults.py#L95>

    Args:
        job: The job object

    Returns:
        The bash code to be inserted
    """


@plugin.spec(result=SimplugResult.ALL_AVAILS)
def on_jobcmd_end(job: Job) -> str:
    """When the job command finishes and after the postscript is run

    This should return a piece of bash code to be inserted in the wrapped job
    script (template), which is a python template string, with the following
    variables available: `status` and `job`. `status` is the class `JobStatus` from
    `xqute.defaults.py` and `job` is the `Job` instance.

    The bash variable `$rc` is accessible in the context, which is the return code
    of the job command.

    For multiple plugins, the code will be inserted in the order of the plugin priority.

    The code will replace the `#![jobcmd_end]` placeholder in the wrapped job script.
    See also <https://github.com/pwwang/xqute/blob/master/xqute/defaults.py#L95>

    Args:
        job: The job object

    Returns:
        The bash code to be inserted
    """


class PipenMainPlugin:
    """The builtin core plugin, used to update the progress bar and
    cache the job"""

    name = "core"
    # The priority is set to -1000 to make sure it is the first plugin
    # to be called
    priority = -1000

    @plugin.impl
    def on_proc_shutdown(proc: Proc, sig: signal.Signals):  # type: ignore[misc]
        """When a process is shutting down"""
        if sig:  # pragma: no cover
            proc.log(
                "warning",
                "Got signal %r, trying a graceful shutdown ...",
                sig.name,
            )

    @plugin.impl
    async def on_job_init(job: Job):  # type: ignore[misc]
        """Update the progress bar when a job is submitted"""
        job.proc.pbar.update_job_inited()

    @plugin.impl
    async def on_job_queued(job: Job):  # type: ignore[misc]
        """Update the progress bar when a job is submitted"""
        job.proc.pbar.update_job_queued()

    @plugin.impl
    async def on_job_started(job: Job):  # type: ignore[misc]
        """Update the progress bar when a job starts to run"""
        job.proc.pbar.update_job_running()

    @plugin.impl
    async def on_job_submitted(job: Job):  # type: ignore[misc]
        """Update the progress bar when a job is submitted"""
        job.proc.pbar.update_job_submitted()

    @plugin.impl
    async def on_job_cached(job: Job):  # type: ignore[misc]
        """Update the progress bar when a job is cached"""
        job.proc.pbar.update_job_queued()
        job.proc.pbar.update_job_submitted()
        job.proc.pbar.update_job_running()
        job.proc.pbar.update_job_succeeded(cached=True)
        await job.set_status(JobStatus.FINISHED)

    @plugin.impl
    async def on_job_succeeded(job: Job):  # type: ignore[misc]
        """Cache the job and update the progress bar when a job is succeeded"""
        # now the returncode is 0, however, we need to check if output files
        # have been created or not, this makes sure job.cache not fail
        for outkey, outtype in job._output_types.items():
            if outtype == ProcOutputType.VAR:
                continue

            path = job.output[outkey].spec
            output_exists = await path.a_exists()
            if outtype == ProcOutputType.DIR:
                is_empty = True
                async for _ in path.a_iterdir():
                    is_empty = False
                    break

                output_exists = output_exists and (not is_empty)

            if not output_exists:
                await job.set_status(JobStatus.FAILED)
                job.proc.pbar.update_job_failed()
                try:
                    # in case the stderr file in the cloud not being synced yet
                    stderr = await job.stderr_file.a_read_text()
                except Exception:  # pragma: no cover
                    stderr = ""

                stderr = f"{stderr}\n\nOutput {outtype} {outkey!r} is not generated"
                if outtype == ProcOutputType.DIR:
                    stderr += " or is empty."
                else:
                    stderr += "."

                await job.stderr_file.a_write_text(stderr)
                break
        else:
            await job.cache()
            job.proc.pbar.update_job_succeeded()

    @plugin.impl
    async def on_job_failed(job: Job):  # type: ignore[misc]
        """Update the progress bar when a job is failed"""
        job.proc.pbar.update_job_failed()
        if job._error_retry and job.trial_count < job._num_retries:  # pragma: no cover
            job.log("debug", "Retrying #%s", job.trial_count + 1)
            job.proc.pbar.update_job_retrying()

    @plugin.impl
    async def on_job_killed(job: Job):  # type: ignore[misc]
        """Update the status of a killed job"""
        # instead of FINISHED to force the whole pipeline to quit
        await job.set_status(JobStatus.FAILED)  # pragma: no cover


plugin.register(PipenMainPlugin)

xqute_plugin = Simplug("xqute")


class XqutePipenPlugin:
    """The plugin for xqute working as proxy for pipen plugin hooks"""

    name = "xqute.pipen"

    @xqute_plugin.impl
    def on_shutdown(xqute: Xqute, sig: signal.Signals):  # type: ignore[misc]
        """When a process is shutting down"""
        return plugin.hooks.on_proc_shutdown(xqute.proc, sig)

    # @xqute_plugin.impl
    # async def on_job_init(scheduler: Scheduler, job: Job):
    #     """When a job is initialized"""
    #     await plugin.hooks.on_job_init(job)

    @xqute_plugin.impl
    async def on_job_queued(scheduler: Scheduler, job: Job):  # type: ignore[misc]
        """When a job is queued"""
        await plugin.hooks.on_job_queued(job)

    @xqute_plugin.impl
    async def on_job_submitting(scheduler: Scheduler, job: Job):  # type: ignore[misc]
        """When a job is being submitted"""
        return await plugin.hooks.on_job_submitting(job)

    @xqute_plugin.impl
    async def on_job_submitted(scheduler: Scheduler, job: Job):  # type: ignore[misc]
        """When a job is submitted"""
        await plugin.hooks.on_job_submitted(job)

    @xqute_plugin.impl
    async def on_job_started(scheduler: Scheduler, job: Job):  # type: ignore[misc]
        """When a job starts to run"""
        await plugin.hooks.on_job_started(job)

    @xqute_plugin.impl
    async def on_job_polling(  # type: ignore[misc]
        scheduler: Scheduler,
        job: Job,
        counter: int,
    ):
        """When a job starts to run"""
        await plugin.hooks.on_job_polling(job, counter)

    @xqute_plugin.impl
    async def on_job_killing(scheduler: Scheduler, job: Job):  # type: ignore[misc]
        """When a job is being killed"""
        return await plugin.hooks.on_job_killing(job)  # pragma: no cover

    @xqute_plugin.impl
    async def on_job_killed(scheduler: Scheduler, job: Job):  # type: ignore[misc]
        """When a job is killed"""
        await plugin.hooks.on_job_killed(job)  # pragma: no cover

    @xqute_plugin.impl
    async def on_job_succeeded(scheduler: Scheduler, job: Job):  # type: ignore[misc]
        """When a job is succeeded"""
        await plugin.hooks.on_job_succeeded(job)

    @xqute_plugin.impl
    async def on_job_failed(scheduler: Scheduler, job: Job):  # type: ignore[misc]
        """When a job is failed"""
        await plugin.hooks.on_job_failed(job)

    @xqute_plugin.impl
    def on_jobcmd_init(scheduler: Scheduler, job: Job):  # type: ignore[misc]
        """When the job command wrapper script is initialized"""
        codes = plugin.hooks.on_jobcmd_init(job)
        if not codes:
            return None
        return "\n\n".join(codes)

    @xqute_plugin.impl
    def on_jobcmd_prep(scheduler: Scheduler, job: Job):  # type: ignore[misc]
        """When the job command is about to be run"""
        codes = plugin.hooks.on_jobcmd_prep(job)
        if not codes:
            return None
        return "\n\n".join(codes)

    @xqute_plugin.impl
    def on_jobcmd_end(scheduler: Scheduler, job: Job):  # type: ignore[misc]
        """When the job command finishes"""
        codes = plugin.hooks.on_jobcmd_end(job)
        if not codes:
            return None
        return "\n\n".join(codes)


xqute_plugin.register(XqutePipenPlugin)



"""Print help for commands"""

from __future__ import annotations
from typing import TYPE_CHECKING

from rich import print

from ._hooks import CLIPlugin

if TYPE_CHECKING:
    from argparse import Namespace

__all__ = ("CLIVersionPlugin",)


class CLIVersionPlugin(CLIPlugin):
    """Print versions of pipen and its dependencies"""

    name = "version"

    def exec_command(self, args: Namespace) -> None:
        """Run the command"""
        import sys
        from importlib.metadata import version
        from .. import __version__

        versions = {"python": sys.version, "pipen": __version__}

        for pkg in (
            "liquidpy",
            "pandas",
            "enlighten",
            "argx",
            "xqute",
            "python-simpleconf",
            "pipda",
            "varname",
        ):
            versions[pkg] = version(pkg)

        keylen = max(map(len, versions))
        for key in versions:
            ver = versions[key]
            verlines = ver.splitlines()
            print(f"{key.ljust(keylen)}: {verlines.pop(0)}")
            for verline in verlines:  # pragma: no cover
                print(f"{' ' * keylen}  {verline}")



"""Provide Cli class"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

from simplug import Simplug

from ..defaults import CLI_ENTRY_GROUP

if TYPE_CHECKING:
    from argx import ArgumentParser
    from argparse import Namespace

cli_plugin = Simplug(CLI_ENTRY_GROUP)


class CLIPlugin(ABC):
    """The abc for cli plugin"""

    def __init__(
        self,
        parser: ArgumentParser,
        subparser: ArgumentParser,
    ) -> None:
        self.parser = parser
        self.subparser = subparser

    @property
    @abstractmethod
    def name(self) -> str:
        """The name/command of this plugin"""

    def parse_args(
        self,
        known_parsed: Namespace,
        unparsed_argv: list[str],
    ) -> Namespace:
        """Define arguments for the command"""
        if unparsed_argv:
            # Let parser raise error for unknown args
            return self.parser.parse_args()

        return known_parsed

    @abstractmethod
    def exec_command(self, args: Namespace) -> None:
        """Execute the command"""


class AsyncCLIPlugin(ABC):
    """The abc for cli plugin with async command execution"""

    def __init__(
        self,
        parser: ArgumentParser,
        subparser: ArgumentParser,
    ) -> None:
        self.parser = parser
        self.subparser = subparser

    async def post_init(self) -> None:
        """Async post init function called after all plugins are loaded"""
        pass

    @property
    @abstractmethod
    def name(self) -> str:
        """The name/command of this plugin"""

    async def parse_args(
        self,
        known_parsed: Namespace,
        unparsed_argv: list[str],
    ) -> Namespace:
        """Define arguments for the command"""
        if unparsed_argv:  # pragma: no cover
            # Let parser raise error for unknown args
            return self.parser.parse_args()

        return known_parsed

    @abstractmethod
    async def exec_command(self, args: Namespace) -> None:
        """Execute the command"""



"""Template adaptor for pipen"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Mapping, Type

from liquid import Liquid

from .defaults import TEMPLATE_ENTRY_GROUP
from .exceptions import NoSuchTemplateEngineError, WrongTemplateEngineTypeError
from .utils import is_subclass, load_entrypoints

__all__ = [
    "Template",
    "TemplateLiquid",
    "TemplateJinja2",
    "get_template_engine",
]


class Template(ABC):
    """Base class wrapper to wrap template for pipen"""

    __slots__ = ("engine",)

    def __init__(
        self,
        source: Any,
        **kwargs: Any,
    ):
        """Template construct"""
        self.engine: Any = None

    def render(self, data: Mapping[str, Any] = None) -> str:
        """
        Render the template
        @parmas:
            data (dict): The data used to render
        """
        return self._render(data or {})

    @abstractmethod
    def _render(self, data: Mapping[str, Any]) -> str:
        """Implement rendering"""


class TemplateLiquid(Template):
    """Liquidpy template wrapper."""

    name = "liquid"

    def __init__(
        self,
        source: Any,
        **kwargs: Any,
    ):
        """Initiate the engine with source and envs

        Args:
            source: The souce text
            envs: The env data
            **kwargs: Other arguments for Liquid
        """
        super().__init__(source)
        self.engine = Liquid(
            source,
            from_file=False,
            mode="wild",
            **kwargs,
        )

    def _render(self, data: Mapping[str, Any]) -> str:
        """Render the template

        Args:
            data: The data used for rendering

        Returns
            The rendered string
        """
        return self.engine.render(data)


class TemplateJinja2(Template):
    """Jinja2 template wrapper"""

    name = "jinja2"

    def __init__(
        self,
        source: Any,
        **kwargs: Any,
    ):
        """Initiate the engine with source and envs

        Args:
            source: The souce text
            envs: The env data
            **kwargs: Other arguments for jinja2.Template
        """
        import jinja2

        super().__init__(source)
        filters = kwargs.pop("filters", {})
        envs = kwargs.pop("globals", {})
        filters = kwargs.pop("filters", {})
        self.engine = jinja2.Template(source, **kwargs)
        self.engine.globals.update(envs)
        self.engine.environment.filters.update(filters)

    def _render(self, data: Mapping[str, Any]) -> str:
        """Render the template

        Args:
            data: The data used for rendering

        Retuens:
            The rendered string
        """
        return self.engine.render(data)


def get_template_engine(template: str | Type[Template]) -> Type[Template]:
    """Get the template engine by name or the template engine itself

    Args:
        template: The name of the template engine or the template engine itself

    Returns:
        The template engine
    """
    if is_subclass(template, Template):
        return template  # type: ignore

    if template == "liquid":
        return TemplateLiquid

    if template == "jinja2":
        return TemplateJinja2

    for name, obj in load_entrypoints(TEMPLATE_ENTRY_GROUP):  # pragma: no cover
        if name == template:
            if not is_subclass(obj, Template):
                raise WrongTemplateEngineTypeError(
                    "Template engine should be a subclass of "
                    "pipen.templates.Template."
                )
            return obj

    raise NoSuchTemplateEngineError(str(template))



"""CLI main entrance"""

import re
import asyncio
import importlib
from pathlib import Path

from argx import ArgumentParser

from ._hooks import cli_plugin, AsyncCLIPlugin
from ..version import __version__

parser = ArgumentParser(
    prog="pipen",
    description=f"CLI Tool for pipen v{__version__}",
)


def load_builtin_clis() -> None:
    """Load builtin cli plugins in this directory"""
    for clifile in Path(__file__).parent.glob("*.py"):
        if clifile.stem.startswith("_"):
            continue
        cli = importlib.import_module(f".{clifile.stem}", __package__)
        plg = getattr(cli, cli.__all__[0])
        cli_plugin.register(plg)


async def _main() -> None:
    """Main function of pipen CLI"""
    cli_plugin.load_entrypoints()
    # builtin plugins have the highest priority
    # so they are loaded later to override the entrypoints
    load_builtin_clis()

    plugin_names = sorted(
        cli_plugin.get_enabled_plugin_names(),
        key=lambda cmd: 999 if cmd == "help" else 0,
    )
    plugins = {}
    for name in plugin_names:
        plg = cli_plugin.get_plugin(name, raw=True)

        docstr = plg.__doc__
        if docstr is not None:
            docstr = docstr.strip()

        subparser = parser.add_command(
            plg.name,
            help=(
                None if docstr is None else re.sub(r"\s+", " ", docstr.splitlines()[0])
            ),
            description=docstr,
        )
        plugins[plg.name] = plg(parser, subparser)

    known_parsed, unparsed_argv = parser.parse_known_args()
    plugin = plugins[known_parsed.COMMAND]
    if isinstance(plugin, AsyncCLIPlugin):
        await plugin.post_init()

    parsed = plugin.parse_args(known_parsed, unparsed_argv)
    if isinstance(plugin, AsyncCLIPlugin):
        parsed = await parsed

    exected = plugin.exec_command(parsed)
    if isinstance(plugin, AsyncCLIPlugin):
        await exected


def main() -> None:
    """The main function of pipen CLI"""
    asyncio.run(_main())



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pipen', 'pipen.cli']

package_data = \
{'': ['*']}

install_requires = \
['argx>=0.4,<0.5',
 'enlighten>=1,<2',
 'liquidpy>=0.8,<0.9',
 'pandas>=2.0,<3.0',
 'pipda>=0.13,<0.14',
 'python-simpleconf[toml]>=0.9,<0.10',
 'varname>=0.15,<0.16',
 'xqute>=2.0,<3.0']

entry_points = \
{'console_scripts': ['pipen = pipen.cli:main']}

setup_kwargs = {
    'name': 'pipen',
    'version': '1.1.12',
    'description': 'A pipeline framework for python',
    'long_description': '<div align="center">\n    <img src="./docs/rtd-logo.png" width="320px" alt="pipen logo showing a stylized pipeline diagram with interconnected nodes">\n\n**A pipeline framework for python**\n\n</div>\n\n______________________________________________________________________\n\n[![Pypi][6]][7] [![Github][8]][9] ![Building][10] [![Docs and API][11]][1] [![Codacy][12]][13] [![Codacy coverage][14]][13] [![Deps][5]][23]\n\n[Documentation][1] | [ChangeLog][2] | [Examples][3] | [API][4]\n\n## Why pipen?\n\n**pipen** is designed for data scientists, bioinformaticians, and researchers who need to create reproducible, scalable computational pipelines without the complexity of traditional workflow systems.\n\n### Target Audience\n\n- **Data Scientists**: Process large datasets with automatic parallelization and caching\n- **Bioinformaticians**: Build reproducible analysis pipelines for genomics data\n- **Researchers**: Create transparent, reproducible workflows for computational research\n- **DevOps Engineers**: Orchestrate batch jobs across different schedulers (SLURM, SGE, Google Cloud)\n\n### Key Benefits\n\n**1. Zero Configuration**\n- Get started immediately with sensible defaults\n- Configure only what you need, when you need it\n- Profile-based configuration for different environments\n\n**2. Reproducibility Built-In**\n- Automatic job caching based on input/output signatures\n- Full audit trail of pipeline runs and parameters\n- Dependency tracking ensures processes run in correct order\n\n**3. Flexible Scheduling**\n- Run locally for development\n- Scale to HPC clusters (SLURM, SGE)\n- Deploy to cloud (Google Cloud Batch, SSH)\n- Run in containers for reproducibility\n\n**4. Developer-Friendly**\n- Define pipelines as Python classes\n- Use familiar Python syntax and tools\n- Extensible plugin system for custom functionality\n- Rich, informative logging and progress tracking\n\n**5. Data Flow Management**\n- Automatic data passing between pipeline stages\n- Support for files, directories, and in-memory data\n- Built-in operations for transforming and aggregating data\n\n### Comparison with Alternatives\n\n| Feature | pipen | Snakemake | Nextflow | Airflow |\n|----------|---------|------------|-----------|----------|\n| **Target Audience** | Data Scientists, Bioinformaticians, Researchers, DevOps | Bioinformaticians | Bioinformaticians | Data Engineers |\n| **Learning Curve** | Low | Medium | High | High |\n| **Python Integration** | Native | Limited | Limited | Native |\n| **Scheduler Support** | 6+ (Local, SGE, SLURM, SSH, Container, Gbatch) | Limited | Limited | Plugin-based |\n| **Caching** | Built-in, automatic | Manual | Manual | Plugin-based |\n| **Cloud Native Support** | Yes (Google Cloud Batch) | Partial | Yes | Yes |\n| **Interactive Debugging** | Yes | Limited | No | No |\n| **Easy to Use** | Define pipelines as Python classes, familiar syntax | Workflow DSL, separate config files | DAG definition in Python, complex UI |\n| **Zero Configuration** | Sensible defaults, configure only what needed | Many configuration options | Heavy configuration required | Complex setup |\n| **Nice Logging** | Rich, informative, color-coded, progress bars | Text-based | Text-based | Basic logging |\n| **Highly Extensible** | Simple plugin system, hook-based | Custom rules/scripts | Custom operators | Custom operators/providers |\n| **Data Flow Management** | Built-in channel operations (expand_dir, collapse_files) | Manual handling | Channel system | XCom system |\n| **Reproducibility** | Built-in caching, full audit trail | Manual | Versioned containers | DAG versioning |\n| **Flexible Scheduling** | Switch schedulers without code changes | Config-based | Config-based | Config-based |\n\n## Installation\n\n```bash\npip install -U pipen\n```\n\n## Quickstart\n\n`example.py`\n\n```python\nfrom pipen import Proc, Pipen, run\n\nclass P1(Proc):\n    """Sort input file"""\n    input = "infile"\n    input_data = ["/tmp/data.txt"]\n    output = "outfile:file:intermediate.txt"\n    script = "cat {{in.infile}} | sort > {{out.outfile}}"\n\nclass P2(Proc):\n    """Paste line number"""\n    requires = P1\n    input = "infile:file"\n    output = "outfile:file:result.txt"\n    script = "paste <(seq 1 3) {{in.infile}} > {{out.outfile}}"\n\n# class MyPipeline(Pipen):\n#     starts = P1\n\nif __name__ == "__main__":\n    # MyPipeline().run()\n    run("MyPipeline", starts=P1)\n```\n\n```shell\n> echo -e "3\\n2\\n1" > /tmp/data.txt\n> python example.py\n```\n\n```log\n04-17 16:19:35 I core                   _____________________________________   __\n04-17 16:19:35 I core                   ___  __ \\___  _/__  __ \\__  ____/__  | / /\n04-17 16:19:35 I core                   __  /_/ /__  / __  /_/ /_  __/  __   |/ /\n04-17 16:19:35 I core                   _  ____/__/ /  _  ____/_  /___  _  /|  /\n04-17 16:19:35 I core                   /_/     /___/  /_/     /_____/  /_/ |_/\n04-17 16:19:35 I core\n04-17 16:19:35 I core                               version: 1.1.8\n04-17 16:19:35 I core\n04-17 16:19:35 I core    ╔═══════════════════════════ MYPIPELINE ════════════════════════════╗\n04-17 16:19:35 I core    ║ My pipeline                                                       ║\n04-17 16:19:35 I core    ╚═══════════════════════════════════════════════════════════════════╝\n04-17 16:19:35 I core    plugins         : verbose v1.1.1\n04-17 16:19:35 I core    # procs         : 2\n04-17 16:19:35 I core    profile         : default\n04-17 16:19:35 I core    outdir          : /path/to/cwd/MyPipeline-output\n04-17 16:19:35 I core    cache           : True\n04-17 16:19:35 I core    dirsig          : 1\n04-17 16:19:35 I core    error_strategy  : ignore\n04-17 16:19:35 I core    forks           : 1\n04-17 16:19:35 I core    lang            : bash\n04-17 16:19:35 I core    loglevel        : info\n04-17 16:19:35 I core    num_retries     : 3\n04-17 16:19:35 I core    scheduler       : local\n04-17 16:19:35 I core    submission_batch: 8\n04-17 16:19:35 I core    template        : liquid\n04-17 16:19:35 I core    workdir         : /path/to/cwd/.pipen/MyPipeline\n04-17 16:19:35 I core    plugin_opts     :\n04-17 16:19:35 I core    template_opts   : filters={\'realpath\': <function realpath at 0x7fc3eba12...\n04-17 16:19:35 I core                    : globals={\'realpath\': <function realpath at 0x7fc3eba12...\n04-17 16:19:35 I core    Initializing plugins ...\n04-17 16:19:36 I core\n04-17 16:19:36 I core    ╭─────────────────────────────── P1 ────────────────────────────────╮\n04-17 16:19:36 I core    │ Sort input file                                                   │\n04-17 16:19:36 I core    ╰───────────────────────────────────────────────────────────────────╯\n04-17 16:19:36 I core    P1: Workdir: \'/path/to/cwd/.pipen/MyPipeline/P1\'\n04-17 16:19:36 I core    P1: <<< [START]\n04-17 16:19:36 I core    P1: >>> [\'P2\']\n04-17 16:19:36 I verbose P1: in.infile: /tmp/data.txt\n04-17 16:19:36 I verbose P1: out.outfile: /path/to/cwd/.pipen/MyPipeline/P1/0/output/intermediate.txt\n04-17 16:19:38 I verbose P1: Time elapsed: 00:00:02.051s\n04-17 16:19:38 I core\n04-17 16:19:38 I core    ╭═══════════════════════════════ P2 ════════════════════════════════╮\n04-17 16:19:38 I core    ║ Paste line number                                                 ║\n04-17 16:19:38 I core    ╰═══════════════════════════════════════════════════════════════════╯\n04-17 16:19:38 I core    P2: Workdir: \'/path/to/cwd/.pipen/MyPipeline/P2\'\n04-17 16:19:38 I core    P2: <<< [\'P1\']\n04-17 16:19:38 I core    P2: >>> [END]\n04-17 16:19:38 I verbose P2: in.infile: /path/to/cwd/.pipen/MyPipeline/P1/0/output/intermediate.txt\n04-17 16:19:38 I verbose P2: out.outfile: /path/to/cwd/MyPipeline-output/P2/result.txt\n04-17 16:19:41 I verbose P2: Time elapsed: 00:00:02.051s\n04-17 16:19:41 I core\n\n             MYPIPELINE: 100%|██████████████████████████████| 2/2 [00:06<00:00, 0.35 procs/s]\n```\n\n```shell\n> cat ./MyPipeline-output/P2/result.txt\n1       1\n2       2\n3       3\n```\n\n## Examples\n\nSee more examples at `examples/` and a more realcase example at:\n\n<https://github.com/pwwang/pipen-report/tree/master/example>\n\n## Plugin gallery\n\nPlugins make `pipen` even better.\n\n- [`pipen-annotate`][26]: Use docstring to annotate pipen processes\n- [`pipen-args`][19]: Command line argument parser for pipen\n- [`pipen-board`][27]: Visualize configuration and running of pipen pipelines on the web\n- [`pipen-diagram`][18]: Draw pipeline diagrams for pipen\n- [`pipen-dry`][20]: Dry runner for pipen pipelines\n- [`pipen-filters`][17]: Add a set of useful filters for pipen templates.\n- [`pipen-lock`][25]: Process lock for pipen to prevent multiple runs at the same time.\n- [`pipen-log2file`][28]: Save running logs to file for pipen\n- [`pipen-poplog`][30]: Populate logs from jobs to running log of the pipeline\n- [`pipen-report`][16]: Generate report for pipen\n- [`pipen-runinfo`][29]: Save running information to file for pipen\n- [`pipen-verbose`][15]: Add verbosal information in logs for pipen.\n- [`pipen-gcs`][32]: A plugin for pipen to handle files in Google Cloud Storage.\n- [`pipen-deprecated`][34]: A pipen plugin to mark processes as deprecated.\n- [`pipen-cli-init`][21]: A pipen CLI plugin to create a pipen project (pipeline)\n- [`pipen-cli-ref`][31]: Make reference documentation for processes\n- [`pipen-cli-require`][24]: A pipen cli plugin check the requirements of a pipeline\n- [`pipen-cli-run`][22]: A pipen cli plugin to run a process or a pipeline\n- [`pipen-cli-gbatch`][33]: A pipen cli plugin to submit pipeline to Google Batch Jobs\n\n[1]: https://pwwang.github.io/pipen\n[2]: https://pwwang.github.io/pipen/CHANGELOG\n[3]: https://pwwang.github.io/pipen/examples\n[4]: https://pwwang.github.io/pipen/api/pipen\n[5]: https://img.shields.io/librariesio/release/pypi/pipen?style=flat-square\n[6]: https://img.shields.io/pypi/v/pipen?style=flat-square\n[7]: https://pypi.org/project/pipen/\n[8]: https://img.shields.io/github/v/tag/pwwang/pipen?style=flat-square\n[9]: https://github.com/pwwang/pipen\n[10]: https://img.shields.io/github/actions/workflow/status/pwwang/pipen/build.yml?style=flat-square\n[11]: https://img.shields.io/github/actions/workflow/status/pwwang/pipen/docs.yml?label=docs&style=flat-square\n[12]: https://img.shields.io/codacy/grade/cf1c6c97e5c4480386a05b42dec10c6e?style=flat-square\n[13]: https://app.codacy.com/gh/pwwang/pipen\n[14]: https://img.shields.io/codacy/coverage/cf1c6c97e5c4480386a05b42dec10c6e?style=flat-square\n[15]: https://github.com/pwwang/pipen-verbose\n[16]: https://github.com/pwwang/pipen-report\n[17]: https://github.com/pwwang/pipen-filters\n[18]: https://github.com/pwwang/pipen-diagram\n[19]: https://github.com/pwwang/pipen-args\n[20]: https://github.com/pwwang/pipen-dry\n[21]: https://github.com/pwwang/pipen-cli-init\n[22]: https://github.com/pwwang/pipen-cli-run\n[23]: https://libraries.io/github/pwwang/pipen#repository_dependencies\n[24]: https://github.com/pwwang/pipen-cli-require\n[25]: https://github.com/pwwang/pipen-lock\n[26]: https://github.com/pwwang/pipen-annotate\n[27]: https://github.com/pwwang/pipen-board\n[28]: https://github.com/pwwang/pipen-log2file\n[29]: https://github.com/pwwang/pipen-runinfo\n[30]: https://github.com/pwwang/pipen-poplog\n[31]: https://github.com/pwwang/pipen-cli-ref\n[32]: https://github.com/pwwang/pipen-gcs\n[33]: https://github.com/pwwang/pipen-cli-gbatch\n[34]: https://github.com/pwwang/pipen-deprecated\n',
    'author': 'pwwang',
    'author_email': 'pwwang@pwwang.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/pwwang/pipen',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



"""Provide JobCaching class that implements caching for jobs"""

from __future__ import annotations
from typing import TYPE_CHECKING
from contextlib import suppress

from diot import Diot
from simpleconf import Config

from .defaults import ProcInputType, ProcOutputType
from .utils import get_mtime, path_is_symlink

if TYPE_CHECKING:
    from xqute.path import SpecPath


class JobCaching:
    """Provide caching functionality of jobs"""

    @property
    def signature_file(self) -> SpecPath:
        """Get the path to the signature file

        Returns:
            The path to the signature file
        """
        return self.metadir / "job.signature.toml"

    async def cache(self) -> None:
        """write signature to signature file"""
        dirsig = (
            self.proc.pipeline.config.dirsig
            if self.proc.dirsig is None
            else self.proc.dirsig
        )
        # Check if mtimes of input is greater than those of output
        try:
            max_mtime = await get_mtime(self.script_file, 0)
        except Exception:  # pragma: no cover
            max_mtime = 0

        # Make self.input serializable
        input_data = {}
        for inkey, intype in self.proc.input.type.items():
            if intype == ProcInputType.VAR:
                input_data[inkey] = self.input[inkey]
                continue

            if intype in (ProcInputType.FILE, ProcInputType.DIR):
                if self.input[inkey] is None:
                    input_data[inkey] = None
                else:
                    input_data[inkey] = str(self.input[inkey].spec)
                    max_mtime = max(
                        max_mtime,
                        await get_mtime(self.input[inkey].spec, dirsig),
                    )

            if intype in (ProcInputType.FILES, ProcInputType.DIRS):
                if self.input[inkey] is None:  # pragma: no cover
                    input_data[inkey] = None
                else:
                    input_data[inkey] = []
                    for file in self.input[inkey]:
                        input_data[inkey].append(str(file.spec))
                        max_mtime = max(max_mtime, await get_mtime(file.spec, dirsig))

        # Make self.output serializable
        output_data = {}
        for outkey, outval in self._output_types.items():
            if outval in (ProcOutputType.FILE, ProcInputType.DIR):
                output_data[outkey] = str(self.output[outkey].spec)
                max_mtime = max(
                    max_mtime,
                    await get_mtime(self.output[outkey].spec, dirsig),
                )
            else:
                output_data[outkey] = self.output[outkey]

        signature = {
            "input": {
                "type": self.proc.input.type,
                "data": input_data,
            },
            "output": {"type": self._output_types, "data": output_data},
            "ctime": float("inf") if max_mtime == 0 else max_mtime,
        }
        async with self.signature_file.a_open("w") as f:
            await f.write(Diot(signature).to_toml())

    async def _clear_output(self) -> None:
        """Clear output if not cached"""
        self.log("debug", "Clearing previous output files.")
        for outkey, outval in self._output_types.items():
            if outval not in (ProcOutputType.FILE, ProcOutputType.DIR):
                continue

            path = self.output[outkey].spec
            if not await path.a_exists() and await path_is_symlink(path):  # dead link
                await path.a_unlink()
            elif await path.a_exists():
                if not await path.a_is_dir():
                    await path.a_unlink()
                else:
                    with suppress(Exception):
                        await path.a_rmtree()
                    # in case rmtree fails anyhow
                    await path.a_mkdir(exist_ok=True)

    async def _check_cached(self) -> bool:
        """Check if the job is cached based on signature

        Returns:
            True if the job is cached otherwise False
        """
        async with self.signature_file.a_open("r") as sf:
            signature = Config.load(await sf.read(), loader="tomls")

        dirsig = (
            self.proc.pipeline.config.dirsig
            if self.proc.dirsig is None
            else self.proc.dirsig
        )

        try:
            # check if inputs/outputs are still the same
            if (
                signature.input.type != self.proc.input.type
                or signature.output.type != self._output_types
            ):
                self.log("debug", "Not cached (input or output types are different)")
                return False

            # check if any script file is newer
            script_mtime = await get_mtime(self.script_file, 0)
            if script_mtime > signature.ctime + 1e-3:
                self.log(
                    "debug",
                    "Not cached (script file is newer: %s > %s)",
                    script_mtime,
                    signature.ctime,
                )
                return False

            # Check if input is different
            for inkey, intype in self.proc.input.type.items():
                sig_indata = signature.input.data.get(inkey)

                if intype == ProcInputType.VAR:
                    if sig_indata != self.input[inkey]:
                        self.log(
                            "debug",
                            "Not cached (input %s:%s is different)",
                            inkey,
                            intype,
                        )
                        return False

                elif int(self.input[inkey] is None) + int(sig_indata is None) == 1:
                    # one is None, the other is not
                    self.log(
                        "debug",
                        "Not cached (input %s:%s is different; "
                        "it is <%s> in signature, but <%s> in data)",
                        inkey,
                        intype,
                        type(sig_indata).__name__,
                        type(self.input[inkey]).__name__,
                    )
                    return False

                elif self.input[inkey] is None and sig_indata is None:
                    continue

                elif intype in (ProcInputType.FILE, ProcInputType.DIR):
                    if sig_indata != str(self.input[inkey].spec):
                        self.log(
                            "debug",
                            "Not cached (input %s:%s is different)",
                            inkey,
                            intype,
                        )
                        return False

                    if (
                        await get_mtime(self.input[inkey].spec, dirsig)
                        > signature.ctime + 1e-3
                    ):
                        self.log(
                            "debug",
                            "Not cached (Input file is newer: %s)",
                            inkey,
                        )
                        return False

                # FILES/DIRS

                # self.input[inkey] can't be None with intype files/dirs
                # elif sig_indata is None:  # both None
                #     continue

                elif not isinstance(sig_indata, list):  # pragma: no cover
                    self.log(
                        "debug",
                        "Not cached (input %s:%s is different, "
                        "%s detected in signature)",
                        inkey,
                        intype,
                        type(sig_indata).__name__,
                    )
                    return False

                else:  # both list
                    if len(sig_indata) != len(self.input[inkey]):  # pragma: no cover
                        self.log(
                            "debug",
                            "Not cached (input %s:%s length is different)",
                            inkey,
                            intype,
                        )
                        return False

                    for i, file in enumerate(self.input[inkey]):
                        if sig_indata[i] != str(file.spec):  # pragma: no cover
                            self.log(
                                "debug",
                                "Not cached (input %s:%s at index %s is different)",
                                inkey,
                                intype,
                                i,
                            )
                            return False

                        if await get_mtime(file.spec, dirsig) > signature.ctime + 1e-3:
                            self.log(
                                "debug",
                                "Not cached (input %s:%s at index %s is newer)",
                                inkey,
                                intype,
                                i,
                            )
                            return False

            # Check if output is different
            for outkey, outtype in self._output_types.items():
                sig_outdata = signature.output.data.get(outkey)
                if outtype == ProcOutputType.VAR:
                    if sig_outdata != self.output[outkey]:  # pragma: no cover
                        self.log(
                            "debug",
                            "Not cached (output %s:%s is different)",
                            outkey,
                            outtype,
                        )
                        return False

                else:  # FILE/DIR
                    if sig_outdata != str(self.output[outkey].spec):  # pragma: no cover
                        self.log(
                            "debug",
                            "Not cached (output %s:%s is different)",
                            outkey,
                            outtype,
                        )
                        return False

                    if not await self.output[outkey].spec.a_exists():
                        self.log(
                            "debug",
                            "Not cached (output %s:%s was removed)",
                            outkey,
                            outtype,
                        )
                        return False

        except Exception as exc:  # pragma: no cover
            # meaning signature is incomplete
            # or any file is deleted
            self.log("debug", "Not cached (%s)", exc)
            return False

        return True

    @property
    async def cached(self) -> bool:
        """Check if a job is cached

        Returns:
            True if the job is cached otherwise False
        """
        out = True
        proc_cache = (
            self.proc.pipeline.config.cache
            if self.proc.cache is None
            else self.proc.cache
        )
        if not proc_cache:
            self.log(
                "debug",
                "Not cached (proc.cache is False)",
            )
            out = False
        elif await self.get_rc() != 0:
            self.log(
                "debug",
                "Not cached (job.rc != 0)",
            )
            out = False
        elif proc_cache == "force":
            try:
                await self.cache()
            except Exception:  # pragma: no cover
                # FileNotFoundError, google.api_core.exceptions.NotFound, etc
                out = False
            else:
                out = True
        else:
            if not await self.signature_file.a_is_file():
                self.log(
                    "debug",
                    "Signature file not found, this is probably an obselete job.",
                )
                await self.cache()

            out = await self._check_cached()

        if not out:
            await self._clear_output()

        return out
